/* 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Linear Least Squares using matrix inverse method example.
 *
 * Grisha Sitkarev, 2011 (c)
 * <sitkarev@unixkomi.ru>
 *
 * How to play:
 *
 * $ gcc -Wall -o lsfit lsfit.c -lgsl -lgslcblas
 * $ ./lsfit
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#define N_ELEMS(x)	(sizeof(x)/sizeof(x[0]))

#define WIDTH	400	/* width of the image */
#define HEIGHT	400	/* height of the image */
#define NCOEF	6	/* number of polynomial coefficients */

#define M	(WIDTH*HEIGHT)

/* This function provides input data. */
void
acquire_data(double *W_data,
	     double *u_data, double *v_data,
	     int height, int width)
{
	double a[NCOEF] = { 0.23, 1.34, 0.22, -0.6, 0.55, 1.28 };
	double b[NCOEF] = { -1.0, 0.66, 0.41, 0.97, -1.27, 1.11 };
	double x, y;
	double *Wp;
	int i, j;

	Wp = W_data;
	y = 0.0;

	fprintf(stdout, "This coefficients will be fitted by Least Square regression.\n");
	
	fprintf(stdout, "a0 = %f a1 = %f a2 = %f a3 = %f a4 = %f a5 = %f\n",
		a[0], a[1], a[2], a[3], a[4], a[5]);

	fprintf(stdout, "b0 = %f b1 = %f b2 = %f b3 = %f b4 = %f b5 = %f\n",
		b[0], b[1], b[2], b[3], b[4], b[5]);

	/*
	 * We are emulating the distortion using the following polynomials:
	 *
	 *   U(x, y) = a0 + a1*x + a2*x^2 + a3*y + a4*y^2 + a5*x*y
	 *   V(x, y) = b0 + b1*x + b2*x^2 + b3*y + b4*y^2 + b5*x*y
	 *
	 * For each [x, y] pair of coordinates starting from [0.0, 0.0] we
	 * calculate corresponding [u, v] pair and store it in the u_data[] and
	 * v_data[] arrays. Later these arrays are used as known parameters in the
	 * Least Square polynomial fit.
	 */

	for (i = 0; i < height; i++) {

		x = 0.0;
	
		for (j = 0; j < width; j++) {

			/* Evaluate polynomials with known coefficients. */
			u_data[i*width + j] = a[0] + x*a[1] + x*x*a[2] +
					  y*a[3] + y*y*a[4] + x*y*a[5];
			
			v_data[i*width + j] = b[0] + x*b[1] + x*x*b[2] +
					  y*b[3] + y*y*b[4] + x*y*b[5];

			/* Calculate matrix W coefficients on the fly. */
			Wp[0] = 1.0;
			Wp[1] = x;
			Wp[2] = x * x;
			Wp[3] = y;
			Wp[4] = y * y;
			Wp[5] = x * y;

			Wp += 6;
			x += 0.25;
		}

		y += 0.25;
	}
}

int
main()
{
	/* input data */
	double u_data[M], v_data[M];
	double W_data[NCOEF*M];
	/* intermediate */
	double inv_data[NCOEF*NCOEF], tmp_data[NCOEF*NCOEF];
	double lu_data[NCOEF*NCOEF];
	double a_data[NCOEF], b_data[NCOEF];
	double tmpv_data[NCOEF];
	gsl_vector_view a, b, u, v, vtmp;
	gsl_matrix_view W, inv, lu, tmp;
	gsl_permutation *p;
	int h, s, w;

	w = WIDTH;	/* width of the fit mesh */
	h = HEIGHT;	/* height of the fit mesh */

	acquire_data(W_data, u_data, v_data, h, w);

	p = gsl_permutation_alloc(NCOEF);

	/* 
	 * Solve overdetermined matrix equations Wa = u, Wb = v using Linear Least
	 * Square method.
	 */

	/* Create matrix W from array data. */
	W = gsl_matrix_view_array(W_data, M, NCOEF);

	/* Create vectors u and v from array data. */
	u = gsl_vector_view_array(u_data, M);
	v = gsl_vector_view_array(v_data, M);

	/* Create vectors a and b. */
	a = gsl_vector_view_array(a_data, NCOEF);
	b = gsl_vector_view_array(b_data, NCOEF);
	vtmp = gsl_vector_view_array(tmpv_data, NCOEF);

	/* Find inverse of W^T*W. */
	tmp = gsl_matrix_view_array(tmp_data, NCOEF, NCOEF);
	lu  = gsl_matrix_view_array(lu_data, NCOEF, NCOEF);
	inv = gsl_matrix_view_array(inv_data, NCOEF, NCOEF);

	/* tmp = W^T * W. */
	gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0,
		       &W.matrix, &W.matrix,
		       0.0, &tmp.matrix);

	/* Get LU decomposition of (W^T * W) in lu. */
	gsl_matrix_memcpy(&lu.matrix, &tmp.matrix);
	gsl_linalg_LU_decomp(&lu.matrix, p, &s);

	/* Obtain (W^T * W)^-1 in inv. */
	gsl_linalg_LU_invert(&lu.matrix, p, &inv.matrix);

	/* tmp = W^T * u. */
	gsl_blas_dgemv(CblasTrans, 1.0, &W.matrix, 
		       &u.vector, 0.0, &vtmp.vector);

	/* Solve for vector u and put coefficients to a. */
	gsl_blas_dgemv(CblasNoTrans, 1.0, &inv.matrix,
		       &vtmp.vector, 0.0, &a.vector);

	fprintf(stderr, "Vector a ===>\n");
	gsl_vector_fprintf(stdout, &a.vector, "%g");

	/* Solve for vector v and put coefficients to b. */ 
	gsl_blas_dgemv(CblasTrans, 1.0, &W.matrix,
		       &v.vector, 0.0, &vtmp.vector);

	gsl_blas_dgemv(CblasNoTrans, 1.0, &inv.matrix,
		       &vtmp.vector, 0.0, &b.vector);

	fprintf(stderr, "Vector b ===>\n");
	gsl_vector_fprintf(stdout, &b.vector, "%g");

	return 0;
}

